/**
 * DataIn, text input window
 * $Id: DataIn.java,v 1.30 2004/07/28 03:13:05 jeffnik Exp $
 */

/* JamochaMUD, a Muck/Mud client program
 * Copyright (C) 1998-2005  Jeff Robinson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package anecho.JamochaMUD;

import java.awt.*;
import java.awt.event.*;

import java.util.Vector;

import anecho.extranet.event.TelnetEvent;
import anecho.extranet.event.TelnetEventListener;

import anecho.gui.SyncFrame;
// import anecho.gui.SyncFrameGroup;
import anecho.gui.ResReader;

/**
 * This is the input window where the user may type to send info
 * to the MUCK/MUD.  In addition, if features scrolling of long messages
 * and a 10 command 'recall' feature implemented by a right mouse-click.
 * @version $Id: DataIn.java,v 1.30 2004/07/28 03:13:05 jeffnik Exp $
 * @author Jeff Robinson
 */
public class DataIn extends SyncFrame implements ActionListener, KeyListener, MouseListener, WindowListener, TelnetEventListener{
    
    private PopupMenu backTrack;
    // private String typedLine; //, bCPopup[];
    private String backComm[];
    // private TextArea dataText;
    private Component dataText, dataInPane;
    private JMConfig settings;
    private CHandler connHandler;
    private JMTFCommands tfCommand;
    private JMTFKeys tfKeys;
    private int limit = 10; // Number of history entries allowed
    private boolean useSwing;
    // private boolean antiAlias = false;
    private static final boolean DEBUG = false;
    private ActionListener pml; // Pop-up menu listener
    private Vector historyV;
    
    public DataIn(JMConfig mainSettings){
        
        super("JamochaMUD");
        
        this.settings = mainSettings;
        this.connHandler = settings.getConnectionHandler();

        historyV = new Vector();
        
        tfCommand = new JMTFCommands(mainSettings);
        tfKeys = JMTFKeys.getInstance();        // Trying out the JMTFKeys singleton
        
        addWindowListener(this);
        // pml = new ActionListener();  // An ActionListener to be used with the pop-up menu
        
        // Set Gridbag layout
        // GridBagLayout dILayout = new GridBagLayout();
        GridBagConstraints constraints = new GridBagConstraints();
        
        backTrack = new PopupMenu();
        backComm = new String[limit];
        
        // Add the text field to the box.
        constraints.gridwidth = 1;
        constraints.gridheight = 1;
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.fill = GridBagConstraints.BOTH;
        
        useSwing = settings.getJMboolean(JMConfig.USESWINGENTRY);
        
        if (useSwing) {
            System.out.println("Swing-based Text entry initialising...");
            // Use a Swing component instead
            // dataText = new anecho.gui.JMSwingEntry("", 3, 70);
            dataText = new anecho.gui.JMSwingEntry("", 70, 3);
            ((anecho.gui.JMSwingEntry)dataText).setLineWrap(true);
            ((anecho.gui.JMSwingEntry)dataText).addKeyListener(this);
            // ((anecho.gui.JMSwingEntry)dataText).add(backTrack); 		// Adds PopupMenu to data entry area
            ((anecho.gui.JMSwingEntry)dataText).addMouseListener(this);	// Adds MouseListener to data entry area
            ((anecho.gui.JMSwingEntry)dataText).setRequestFocusEnabled(true);
            ((anecho.gui.JMSwingEntry)dataText).setAntiAliasing(settings.getJMboolean(JMConfig.ANTIALIAS));
            // ((anecho.gui.JMSwingEntry)dataText).setCaretColor(settings.getJMColor(settings.BACKGROUNDCOLOUR));
            
            dataInPane = new javax.swing.JScrollPane((anecho.gui.JMSwingEntry)dataText);
            ((javax.swing.JScrollPane)dataInPane).setVerticalScrollBarPolicy(javax.swing.JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
            ((javax.swing.JScrollPane)dataInPane).setHorizontalScrollBarPolicy(javax.swing.JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
            // ((anecho.gui.JMSwingEntry)dataText).setRows(3);
            add((javax.swing.JScrollPane)dataInPane);
            if (DEBUG) {
                System.err.println("Done.");
            }
        } else {
            // Standard AWT TextArea
            dataText = (Component)new TextArea("", 3, 70, TextArea.SCROLLBARS_VERTICAL_ONLY);
            ((TextArea)dataText).addKeyListener(this);
            ((TextArea)dataText).add(backTrack); 		// Adds PopupMenu to data entry area
            ((TextArea)dataText).addMouseListener(this);	// Adds MouseListener to data entry area
            add((TextArea)dataText);
        }
        
        backTrack.addActionListener(this);
        pack();
        
        // Sure, this is a bit of work, but we so love our icons, don't we?
        MediaTracker tracker = new MediaTracker(this);
        Image jMUDImage = Toolkit.getDefaultToolkit().getImage("kehza.gif");
        tracker.addImage(jMUDImage, 0);
        try {
            tracker.waitForAll();
        } catch (Exception exc) {
            System.out.println("Icon load error.  Exception " + exc);
        }
        this.setIconImage(jMUDImage);
        
        // Display macroBar
        // dataText.setForeground(settings.getForegroundColour());
        // dataText.setBackground(settings.getBackgroundColour());
        dataText.setForeground(settings.getJMColor(settings.FOREGROUNDCOLOUR));
        dataText.setBackground(settings.getJMColor(settings.BACKGROUNDCOLOUR));
        // dataText.setFont(settings.getFontFace());
        dataText.setFont(settings.getJMFont(settings.FONTFACE));
    }
    
    public void keyPressed(KeyEvent event){
        
        int arg = event.getKeyCode();
        
        if (arg == KeyEvent.VK_ENTER) {
            if (useSwing) {
                JMSendText(((anecho.gui.JMSwingEntry)dataText).getText());
            } else {
                JMSendText(((TextArea)dataText).getText());
            }
            event.consume();
        }
        
        if (arg == KeyEvent.VK_ALT) {
            // Check to see if the user wants to have to ALT key focus on the menu
            // if (settings.getAltFocus()) {
            if (settings.getJMboolean(JMConfig.ALTFOCUS) && settings.getJMboolean(JMConfig.SPLITVIEW)) {
                // (settings.getMainWindowVariable()).requestFocus();
                if (DEBUG) {
                    System.err.println("DataIn.keyPressed() sending focus to MuckMainFrame.");
                }
                (settings.getJMFrame(JMConfig.MUCKMAINFRAME)).requestFocus();
            }
        }
        
        /*
         if (arg == KeyEvent.VK_UP) {
         // Check to see if the cursor is allowed to move up
         if (useSwing) {
         if (((TextArea)dataText).getCaretPosition() <= ((TextArea)dataText).getColumns()) {
         // Caret is on first line, can't move up
         event.consume();
         }
         } else {
         if (((anecho.gui.JMSwingEntry)dataText).getCaretPosition() <= ((anecho.gui.JMSwingEntry)dataText).getColumns()) {
         event.consume();
         }
         }
         }
         */
        
        if ((arg == KeyEvent.VK_X) && (event.isControlDown())) {
            // 'Supervisor' key was called.
            // Minimize all the windows
            HideFramesQuick();
        }
        
        if ((arg == KeyEvent.VK_Z) && (event.isControlDown())) {
            // restore minimized windows
            RestoreMinimizedFrames();
            
        }
        
        if (event.isControlDown()) {
            // Check for any "editor" controls, such as CTRL+SHIFT+ARROW
            if (arg == KeyEvent.VK_LEFT || arg == KeyEvent.VK_RIGHT) {
                // Nothing to see here, let the system handle it!
                // return;
                event.consume();
            } else {
                
                // First, check to see if this is for TinyFugue emulation
                if (settings.getJMboolean(JMConfig.TFKEYEMU) && tfKeys.jmTFKeyStroke(settings, event.getKeyCode())) {
                    event.consume();
                }
                
                // We'll send this to MuckMain, to see if
                // it is supposed to be a menu shortcut
                // MuckMain tempMain = settings.getMainWindowVariable();
                if (settings.getJMboolean(JMConfig.ALTFOCUS) && settings.getJMboolean(JMConfig.SPLITVIEW)) {
                    Frame tempMain = settings.getJMFrame(JMConfig.MUCKMAINFRAME);
                    tempMain.dispatchEvent(event);
                }
                
            }
        }
        
        return;
    }
    
    public void keyReleased(KeyEvent event){}
    
    public void keyTyped(KeyEvent event){
        // We must make one exception, if the user is using TinyFugue emulation
        // so that we do not spool out text if they're releasing the 'pause text' key
        // if (settings.getTFKeyEmu() && event.isControlDown()) {
        if (settings.getJMboolean(settings.TFKEYEMU) && event.isControlDown()) {
            event.consume();
        } else {
            checkPause();
        }
    }
    
    public void actionPerformed(ActionEvent event){
        if (DEBUG) {
            System.err.println("DataIn received ActionEvent " + event);
        }
        
        // String arg = event.getActionCommand();
        
        String cmd = event.getActionCommand();
        
        int cmdInt = 0;
        
        try {
            cmdInt = Integer.parseInt(cmd);
            if (DEBUG) {
                System.err.println("Chosen action command is " + cmd);
            }
            JMSetFromScrollBack(cmdInt);
        } catch (Exception e) {
            if (DEBUG) {
                System.err.println("Problem getting action command value.");
            }
        }
        
        // Determine if a popupmenu item was selected (until a better way is found)
//        if (useSwing) {
//            // Fix Me XXX
//        } else {
//            for (int i = 0; i < backTrack.getItemCount(); i++) {
//                if (arg.equals((backTrack.getItem(i).getLabel()))) {
//                    JMSetFromScrollBack(i);
//                }
//            };
//        }
        return;
    }
    
    // Check for mouse actions
    
    public void mousePressed(MouseEvent evt) {
        checkPopup(evt) ;
    }
    
    public void mouseReleased(MouseEvent evt) {
        checkPopup(evt) ;
    }
    
    public void mouseClicked(MouseEvent evt) {
        // This spools out any paused text
        checkPause();
    }
    
    public void mouseEntered(MouseEvent evt) {
    }
    
    public void mouseExited(MouseEvent evt) {
    }
    
    /**
     * Check for PopupTrigger to display command history
     */
    void checkPopup(MouseEvent mouse) {
        if (mouse.isPopupTrigger()) {
            // ShowPopup(this, mouse.getX(), mouse.getY());
            // ShowPopup(dataText, mouse.getX(), mouse.getY());
            if (DEBUG) {
                System.err.println("Our original component is " + mouse.getComponent());
            }
            
            ShowPopup(mouse.getComponent(), mouse.getX(), mouse.getY());
            mouse.consume();
        }
    }
    
    /** Display the pop-up menu */
    public void ShowPopup(Component origin, int x, int y) {
        Component objOrg = origin;
        int xaxis = x;
        int yaxis = y;
        
        if (DEBUG) {
            System.err.println("DataIn.ShowPopup: origin object: " + objOrg);
        }
        
        // if x and y come in as -1, then we set the popup location from just
        // the frame as this method was probably called by an outside class
        if (origin == null && xaxis == -1 && yaxis == -1) {
            if (DEBUG) {
                System.err.println("DataIn.ShowPopup: Our axi are both -1.");
            }
            // objOrg = this;
            objOrg = dataText;
            xaxis = (int)(this.getSize().width / 2);
            yaxis = (int)(this.getSize().height / 2);
            
            if (DEBUG) {
                System.err.println("Our new objOrg is " + objOrg);
                System.err.println("xaxis: " + xaxis);
                System.err.println("yaxis: " + yaxis);
            }
        }
        
        if (useSwing) {
            javax.swing.JMenuItem tempItem;
            javax.swing.JPopupMenu jpm = new javax.swing.JPopupMenu("Command history");
            String tempName;
//            for (int i = 0; i < backComm.length; i++) {
//                if (backComm[i] != null) {
//                    if (backComm[i].length() > 30) {
//                        // jpm.add(backComm[i].substring(0, 30) + "...");
//                        tempName = backComm[i].substring(0, 30) + "...";
//                    } else {
//                        // jpm.add(backComm[i]);
//                        tempName = backComm[i];
//                    }
//                    
//                    if (DEBUG) {
//                        System.err.println("Adding item: " + tempName);
//                    }
//                    
//                    tempItem = new javax.swing.JMenuItem(tempName);
//                    tempItem.addActionListener(pml);
//                    jpm.add(tempItem);
//                } else {
//                    break;
//                }
//            }
            for (int i = 0; i < historyV.size(); i++) {
                tempName = historyV.elementAt(i).toString();
                tempItem = new javax.swing.JMenuItem(tempName);
                // tempItem.addActionListener(pml);
                tempItem.addActionListener(this);
                tempItem.setActionCommand(i + "");
                jpm.add(tempItem);
            }
            
            // jpm.addPopupMenuListener(javax.swing.event.PopupMenuListener pml);
            // jpm.pack();
            jpm.show(objOrg, xaxis, yaxis);
            
        } else {
            backTrack.removeAll();
            
//            for (int i = 0; i < backComm.length; i++) {
//                if (backComm[i] != null) {
//                    if (backComm[i].length() > 30) {
//                        backTrack.add(backComm[i].substring(0, 30) + "...");
//                    } else {
//                        backTrack.add(backComm[i]);
//                    }
//                } else {
//                    break;
//                }
//            }
            String tempName;
            java.awt.MenuItem tempItem;
            
            for (int i = 0; i < historyV.size(); i++) {
                tempName = historyV.elementAt(i).toString();
                tempItem = new java.awt.MenuItem(tempName);
                tempItem.addActionListener(pml);
                tempItem.setActionCommand(i + "");
                backTrack.add(tempItem);
            }
            
            backTrack.show(objOrg, xaxis, yaxis);
        }
    }
    
    /**
     * Bring the frame into focus so the user may type
     */
    public void windowActivated(WindowEvent event) {
        // First we insert a 6 millisecond kludge to make
        // up for the requestFocus timing problem in Java 1.1.7+
        long startTime = java.lang.System.currentTimeMillis();
        long currentTime = java.lang.System.currentTimeMillis();
        
        while ((currentTime - startTime) < 7) {
            // Just an empty loop
            currentTime = java.lang.System.currentTimeMillis();
        }
        
        // Now request focus
        if (useSwing) {
            System.out.println("We asked for a requestFocus");
            ((anecho.gui.JMSwingEntry)dataText).requestFocus();
        } else {
            ((TextArea)dataText).requestFocus();
        }
        
    }
    
    public void windowClosed(WindowEvent event) {}
    
    public void windowClosing(WindowEvent event) {}
    
    public void windowDeactivated(WindowEvent event) {}
    
    public void windowDeiconified(WindowEvent event) {}
    
    public void windowIconified(WindowEvent event) {
        // record the frame into the hashtable
        // settings.setDataBar(getBounds());
        settings.setJMValue(JMConfig.DATABAR, getBounds());
    }
    
    public void windowOpened(WindowEvent event) {}
    
    // Our customer TelnetEventListener
    public void telnetMessageReceived(TelnetEvent event) {
        // System.out.println("Confirming that DataIn has received: " + event);
        // System.out.println("With our message being: " + event.getMessage());
        // String message = event.getMessage();
        // if (message.equals("IAC DO TELOPT_ECHO")) {
        // dataText.setEchoChar('*');
        // }
        
        // if (message.equals("IAC DONT TELOPT_ECHO")) {
        // dataText.setEchoChar('\u0000');
        // }
    }
    
    /**
     * This is a generic bit to access the
     * ResReader.class, for localization
     * (Multi-language support)
     */
    private String RB(String itemTarget) {
        ResReader reader = new ResReader();
        // return reader.LangString("JamochaMUDBundle", itemTarget);
        return reader.LangString(JMConfig.BUNDLEBASE, itemTarget);
    }
    
    /**
     * Method to hide all the frames.  The 'supervisor' function
     */
    public void HideFramesQuick() {
        // Hide all the frames,
        // the supervisor's here!
        setTitle(".");
        toBack();
        
        // We'll unsync the databar, otherwise bringing it up to the
        // front may result in all the other windows showing up again.  Eeek!
        setSync(false);
    }
    
    /**
     * Restore frames hidden by the 'Supervisor' function
     */
    public void RestoreMinimizedFrames() {
        // Restore the frames that
        // were minimized by the 'supervisor' key
        // Fix this XXX
        // MuckMain.textWindow.setVisible(true);
        
        // Restore frame icons
        // dataBar.setIconImage(MuckConn.jMUDImage);
        // MuckMain.textWindow.setIconImage(MuckConn.jMUDImage);
        
        // Querry the MuckMain **** menu, to
        // see if we should set the Macros visible
        // Fix this XXX
        // if (MuckMain.tWMacro.getState()) {
        // tWMacro is set to 'true', so show the Macro Frame
        // MuckConn.jmMacros.setVisible(true);
        //}
        
        // Fix this XXX
        // if (MuckMain.tWSyncWindowsItem.getState()) {
        // setSync(true);
        // }
        // now reset the title
        // resetTitles();
    }
    
    /**
     * Add an entry to the 'command history' pop-up
     */
    /*
    private void addPopUpEntry(String str) {
        if (backTrack.getItemCount() > 9) {
            backTrack.remove(0);  // Removes first entry to make room
            for(int i = 0; i < backTrack.getItemCount(); i++) {
                backComm[i] = backComm[i + 1];
            }
        }
     
        backComm[backTrack.getItemCount()] = str;
        if (str.length() > 30) {
            backTrack.add(str.substring(0, 30) + "...");
        } else {
            backTrack.add(str);
        }
    }
     */
    
    private void updateEntryList(String str) {
        // int curSize = backComm.length - 1;  // Current number of items in our list
        // int curSize = 0;
        
        historyV.addElement(str);
        
        if (historyV.size() > limit) {
            historyV.removeElementAt(0);
        }
        
//        if (backComm[0] == null) {
//            if (DEBUG) {
//                System.err.println("backComm history list is empty.  Adding: " + str);
//            }
//            backComm[0] = str;
//        } else {
//            
//            int loopLimit = limit -1;
//            for (int i = 0; i < loopLimit; i++) {
//                curSize = i;
//                if (backComm[i + 1] != null) {
//                    backComm[i] = backComm[i + 1];
//                } else {
//                    break;
//                }
//            }
//            backComm[curSize + 1] = str;
//            
//        }
//        
//        //        if (curSize > 0) {
//        //            if (curSize == limit) {
//        //                curSize = limit - 1;
//        //            }
//        //        }
//        
//        // drop the oldest item from our list
////        for (int i = 0; i < curSize; i++) {
////            backComm[i] = backComm[i + 1];
////        }
////        
//        backComm[curSize] = str;
        if (DEBUG) {
            System.err.println("Successfully reached the end of updating history.");
        }
    }
    
    /**
     * This is an &quot;empty" method that allows other
     * classes to request that we send our existing text
     * to the MU*.  This doesn't happen often, but can
     * be a handy ability to have */
    public void JMSendText() {
        if (useSwing) {
            JMSendText(((anecho.gui.JMSwingEntry)dataText).getText());
        } else {
            JMSendText(((TextArea)dataText).getText());
        }
    }
    
    /**
     * Package up the data to be sent to the MU*, parsing off any
     * yucky characters such as 'new lines', etc., also determining
     * via user-preference whether they should be sent in Unicode
     * or standard ASCII format
     */
    public void JMSendText(String outGoing) {
        String sendStr = outGoing;
        // Safety first!
        
        if (connHandler == null) {
            connHandler = settings.getConnectionHandler();
        }
        
        // Now add the command to the Popupmenu
        // addPopUpEntry(sendStr);
        updateEntryList(sendStr);
        
        // Check to see if this is a TinyFugue-like command
        // if (sendStr.startsWith("/") && settings.getJMboolean(settings.TFKEYEMU) && tfCommand.Command(sendStr)) {
        if (!(sendStr.startsWith("/") && settings.getJMboolean(settings.TFKEYEMU) && tfCommand.Command(sendStr))) {
            
            // Send the text to the appropriate MU* via its connection handler
            sendStr = translateMacros(sendStr);
            connHandler.sendText(sendStr);
        }
        
        // If local echo is active, append it to the user's screen as well
        // if (settings.isLocalEchoEnabled()) {
        if (settings.getJMboolean(JMConfig.LOCALECHO)) {
            if (connHandler.getActiveMUEchoState()) {
                if (settings.getJMboolean(JMConfig.USESWING)) {
                    anecho.gui.JMSwingText text;
                    text = connHandler.getActiveMUDSwingText();
                    text.append(RB("localEchoPrepend") + " " + sendStr + '\n');
                } else {
                    anecho.gui.JMText text;
                    text = connHandler.getActiveMUDText();
                    text.append(RB("localEchoPrepend") + " " + sendStr + '\n');
                }
            }
        }
        
        // return the cursor to the starting position and empty the text area
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).setCaretPosition(0);
            ((anecho.gui.JMSwingEntry)dataText).setText("");
        } else {
            ((TextArea)dataText).setCaretPosition(0);
            ((TextArea)dataText).setText("");
        }
        
        
    }
    
    /** Set the current text from the scrollback menu,
     * the item indicated by int pos */
    public void JMSetFromScrollBack(int currentPos) {
        int pos = currentPos;
        
        // If the integer comes in as '-1', that means to use the last string
        if (pos == -1) {
            // pos = backTrack.getItemCount() - 1;
            pos = historyV.size() - 1;
        }
        
        // First, check to see if the integer is valid
//        if (pos > backTrack.getItemCount() || pos < 0) {
//            return;
//        }
        
        if (pos > historyV.size() || pos < 0) {
            return;
        }
        
        String text = historyV.elementAt(pos).toString();
        
        if (useSwing) {
            // ((anecho.gui.JMSwingEntry)dataText).setText(backComm[pos]);
            ((anecho.gui.JMSwingEntry)dataText).setText(text);
            ((anecho.gui.JMSwingEntry)dataText).setCaretPosition(backComm[pos].length());
        } else {
            // Now we'll set the text from the menu item
            // ((TextArea)dataText).setText(backComm[pos]);
            ((TextArea)dataText).setText(text);
            // Explicitly set the cursor location, as this is
            // not automatic on some operating systems.
            ((TextArea)dataText).setCaretPosition(backComm[pos].length());
            
        }
    }
    
    /** Some text methods */
    public void setText(String text) {
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).setText(text);
        } else {
            ((TextArea)dataText).setText(text);
        }
        
    }
    
    public void setFont(Font newFont) {
        
        dataText.setFont(newFont);
    }
    
    public void append(String text) {
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).append(text);
        } else {
            ((TextArea)dataText).append(text);
        }
    }
    
    public String getText() {
        String retText = "";
        if (useSwing) {
            retText = ((anecho.gui.JMSwingEntry)dataText).getText();
        } else {
            retText = ((TextArea)dataText).getText();
        }
        // return ((TextArea)dataText).getText();
        return retText;
    }
    
    public void setCaretPosition(int pos) {
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).setCaretPosition(pos);
        } else {
            ((TextArea)dataText).setCaretPosition(pos);
        }
    }
    
    public int getCaretPosition() {
        int retInt = 0;
        if (useSwing) {
            retInt = ((anecho.gui.JMSwingEntry)dataText).getCaretPosition();
        } else {
            retInt = ((TextArea)dataText).getCaretPosition();
        }
        // return ((TextArea)dataText).getCaretPosition();
        return retInt;
    }
    
    public int getColumns() {
        int retCol = 0;
        if (useSwing) {
            retCol = ((anecho.gui.JMSwingEntry)dataText).getColumns();
        } else {
            retCol = ((TextArea)dataText).getColumns();
        }
        return retCol;
        // return ((TextArea)dataText).getColumns();
    }
    
    public void setForegroundColour(Color fgColour) {
        dataText.setForeground(fgColour);
    }
    
    public void setBackgroundColour(Color bgColour) {
        dataText.setBackground(bgColour);
    }
    
    /** Set the title of this frame on the basis of if the
     * active MU* (the one visible at the moment) is
     * connected or not.  This method differs from the main
     * window's setWindowTitle() method in that it takes
     * the text being paused into consideration as well
     */
    public synchronized void setWindowTitle() {
        // This resets the window's title, depending if the setWindowTitle();
        // client is connected to a MU* or not
        CHandler connection = settings.getConnectionHandler();
        boolean active = connection.isActiveMUDConnected();
        
        if (active) {
            MuSocket mSock = connection.getActiveMUHandle();
            
            if (mSock.isPaused()) {
                this.setTitle(RB("outputPaused"));
            } else {
                this.setTitle(connection.getActiveTitle());
            }
        } else {
            // Connection is inactive, just have program's title
            this.setTitle("JamochaMUD - " + RB("notConnected"));
        }
    }
    
    /** We export our TextArea for use in a &quot;combined"
     * JamochaMUD configuration, where both the input and
     * output are shown in the same frame
     */
    // public TextArea exportText() {
    // public TextArea exportText() {
    public Component exportText() {
        Component retComp;
        this.setVisible(false);
        this.setSync(false);
        
        
        if (useSwing) {
            retComp = dataInPane;
        } else {
            retComp = dataText;
        }
        // We should also remove this frame from the sync-group?
        // return (TextArea)dataText;
        // return dataText;
        return retComp;
    }
    
    /** We are having our TextArea &quot;returned" to us from
     * the main window.  We'll add it back to our frame and
     * then set everything visible again
     */
    public void restoreText() {
        if (useSwing) {
            this.add((anecho.gui.JMSwingEntry)dataText);
        } else {
            this.add((TextArea)dataText);
        }
        // this.add(dataText);
        this.pack();
        
        // Check for sync value to restore us to our previous glory
        // if (settings.getSyncWindows()) {
        if (settings.getJMboolean(JMConfig.SYNCWINDOWS)) {
            this.setSync(true);
        }
        
        this.setVisible(true);
    }
    
    public Component getTextVariable() {
        Component retComp;
        
        if (useSwing) {
            retComp = dataInPane;
        } else {
            retComp = dataText;
        }
        // We should also remove this frame from the sync-group?
        // return (TextArea)dataText;
        // return dataText;
        return retComp;
    }
    
    /** Remove the paused status from JamochaMUD because of some
     * action that we have received
     */
    private void checkPause() {
        CHandler target = settings.getConnectionHandler();
        MuSocket mSock = target.getActiveMUHandle();
        
        if (mSock != null) {
            if (mSock.isPaused()) {
                // This spools out any paused text
                mSock.spoolText();
            }
        }
    }
    
    public void setRows(int rows) {
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).setRows(rows);
        } else {
            ((TextArea)dataText).setRows(rows);
        }
    }
    
    /** Do any required macro replacement in our string */
    private String translateMacros(String macroIn) {
        String input = macroIn;
        boolean process = true;
        int start, end;
        String command, sString, eString, macro;
        
        /** Create a loop to continuously process any (recursive) macros we find */
        // We need to be able to locate endless loops!  Fix this XXX
        while (process) {
            start = input.indexOf("${");
            
            if (start > -1) {
                
                end = input.indexOf("}", start);
                if (end > -1 && end > start) {
                    command = input.substring(start + 2, end);
                    // System.out.println("Our macro is: " + command);
                    
                    if (start > 0) {
                        sString = input.substring(0, start);
                    } else {
                        sString = "";
                    }
                    if (end + 1 < input.length()) {
                        eString = input.substring(end + 1);
                    } else {
                        eString = "";
                    }
                    
                    macro = settings.getVariable(command);
                    // System.out.println("Our macro returns: " + macro);
                    //if (macro != null) {
                    // input = sString + macro + eString;
                    // } else {
                    // input = sString + command + eString;
                    // }
                    
                    if (macro == null) {
                        input = sString + command + eString;
                    } else {
                        input = sString + macro + eString;
                    }
                    
                }
                
            } else {
                // No more macros to process
                process = false;
            }
        }
        
        return input;
    }
    
    /** This lets our other classes bring us into focus */
    public void jmGainFocus() {
        if (useSwing) {
            System.out.println("JMGainFocus()");
            ((anecho.gui.JMSwingEntry)dataText).requestFocus();
            ((anecho.gui.JMSwingEntry)dataText).grabFocus();
        } else {
            ((TextArea)dataText).requestFocus();
        }
    }
    
    // This can probably be removed XXX
    public synchronized void validate() {
        dataText.invalidate();
    }
    
    /** Return the last word typed */
    public String getLastWord() {
        String retString = "";
        
        String inputText = "";
        int cursorPos = 0;
        
        if (useSwing) {
            inputText = ((anecho.gui.JMSwingEntry)dataText).getText();
            cursorPos = ((anecho.gui.JMSwingEntry)dataText).getCaretPosition();
        } else {
            inputText = ((TextArea)dataText).getText();
            cursorPos = ((TextArea)dataText).getCaretPosition();
        }
        
        // System.out.println("inputText: " + inputText);
        //System.out.println("cursorPos: " + cursorPos);
        
        if (cursorPos > 1) {
            // We use cursorPos - 2 as by the time we check, the space is already in place!
            int start = inputText.lastIndexOf(" ", cursorPos - 2);
            if (start < 0) {
                start = 0;
            }
            // System.out.println("Start point: " + start);
            retString = inputText.substring(start, cursorPos);
            // retString = inputText.substring(start, cursorPos - 1);
        }
        
        return retString.trim();
    }
    
    /** Enable or disable anti-aliasing */
    public void setAntiAliasing(boolean state) {
        if (useSwing) {
            ((anecho.gui.JMSwingEntry)dataText).setAntiAliasing(state);
        }
        
    }
    
}
